/**
 * File:   cObject.h
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on March 28, 2008, 10:23 PM (1999)
 *
 * Generic Game-Object with as much common
 * attributes as making sense and as needed
 * for various World-Filtering (Worldsense for Objects).
 */

#ifndef OBJECT_INC
#define OBJECT_INC

class cObject;
class cPad;
class cController;

#include "cWorld.h"
#include <string>

using namespace std;

/**
 * Fundamental function for collision-check.
 */
float collideLineWithAAB(float* start3fv, float* end3fv, float* min3fv, float* max3fv, float* impactpoint3fv, float* impactnormal3fv);

/**
 * Generic Game-Object (Object with Roles) which
 * collects as much common attributes as are needed
 * by the different roles and as needed by World-Filtering
 * (Worldsense for subclasses implementing the roles).
 * This object does again hold as much methods as needed
 * but without (real) implementation (which is done by the subclasses).
 * Nevertheless every method has a empty body or a body returning
 * some default value, so that callers need not care about
 * the kind of the object (roles) too much.
 */
class cObject {
public:
  friend class cWorld;
  friend class cController;
  
  enum Roles {
    // Roles indicating Attributes
    NAMEABLE,
    SOCIALISED,
    TRACEABLE,
    COLLIDEABLE,
    DAMAGEABLE,
    ENTITY,
    
    // Roles indicating Kind of Objects
    MECH,
    GEOSPOT,
    BUILDING,
    TREE,
    
    // Roles indicating belonging to certain Social-Parties
    RED,
    BLUE,
    GREEN,
    YELLOW,
    CIVIL,

    // Roles indicating State
    RELOADING,

    WOUNDED, //  <= 75% hp
    SERIOUS, // <= 50% hp
    CRITICAL, // <= 25% hp
    DEAD, // <= 0% hp, note that a the later include the upper.

    // Special Roles
    HUMANPLAYER,
  };
  
  // BASE
  bool mFragged; // Is this object scheduled to be deleted.
  unsigned long mRoles; // Bitmask, lists the roles this object plays.
  double mSeconds; // Simulated Seconds since spawn, update in animateSelf!

  // NAMEABLE
  string* mName;
  string* mDescription;
  unsigned int mDesignation; // { 0, .., 25 } for { a, ..., z }.

  // SOCIALISED
  unsigned long mAllies; // future: Bitmask, lists Roles which Allies have.
  unsigned long mEnemies; // Bitmask, lists Roles which Enemies play.
  
  // TRACEABLE
  float mAcc[3];
  float mVel[3];
  float mPos[3];
  float mRot[3];
  float mRadius;
  
  enum Parts {
      LEFT,
      RIGHT,
      BODY,
      LEGS,
      MAX_PARTS
  };
  
  // DAMAGEABLE
  float mHp;
  //int* mSplasharmor;
  //int* mBeamarmor;
  //int* mPiercearmor;
  // int mHeat;
  // int mHeatsinks;
  // int mSplasharmor;
  // int mBeamarmor;
  // int mPiercearmor;

  // ENTITY
  cObject* mTarget; // Aim target
  float mDestination[3]; // Movement target
  cPad* mPad; // Human, AI/Remote Pad-Control over the Object.
  cController* mController; // Command Stack takes controll over the pad when enabled.
  
public:
  cObject();
  cObject(cObject* source);
  virtual ~cObject() {}

  // Roles
  bool hasRole(Roles role) { return (mRoles & (1UL << role)); }
  void addRole(Roles role) { mRoles |= (1UL << role); }
  void remRole(Roles role) { mRoles ^= (1UL << role); }

  // Called right after object was spawned into the world.
  virtual void onSpawn() {}
  
  // 
  virtual void onFrag() {}
  
  // Multiplies in the Object's camera matrix
  virtual void multEyeMatrix() {}
  
  virtual void animateSelf(float spf) { mSeconds += spf; }
  virtual void transformSelf() {}
  virtual void drawSolid() {}
  virtual void drawEffect() {}
  virtual void drawHUD() {}
  
  // Receiving damage:
  
  struct rImpact { float pos[3]; float nrm[3]; };
  
  // Deal damage to the object (parts) (if applicable) when possible at
  // intersection with the described cylinder/beam, fallback to a line
  // or other approximation as you wish.
  // all: deal damage to anything intersected not just the nearest part.
  virtual rImpact* damageByBeam(float* a, float* b, float radius, float damage, bool all) { return NULL; }
  
  // Deal damage to object (parts) inside the described sphere.
  // all: deal damage to anything intersected not just the nearest part.
  //virtual rImpact* damageBySphere(float* a, float radius, float damage, bool all) { return NULL; };

  // Deal damage to object (parts) without respect to any location.
  //virtual void damage(float damage) {};
  
  // Related to States:
  
  // Queries
  virtual cObject* enemyNearby() { return NULL; }
    
  // Fuzzy Predicates [0,1]
  virtual float inMeeleRange() { return 0.0f; }
  virtual float inWeaponRange() { return 0.0f; }
  virtual float inTargetRange() { return 0.0f; }

  // Output Actions for Entities
  virtual void do_moveTowards() {}
  virtual void do_moveNear() {}
  virtual void do_aimAt() {}
  virtual void do_fireAt() {}
  virtual void do_idle() {}
  virtual void do_aimFor(cObject* target) {}
};

#endif
